### Fitting the neural networks

mport json
import os
import pickle
from math import sqrt
from pathlib import Path
from pprint import pprint

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import sklearn
import torch
import torch.nn as nn
from scipy import interpolate
from sklearn.ensemble import RandomForestRegressor
from sklearn.metrics import mean_absolute_error, mean_squared_error, r2_score
from sklearn.model_selection import (GridSearchCV, KFold, PredefinedSplit,
                                     RandomizedSearchCV, cross_val_score,
                                     cross_validate, train_test_split)
from skorch import NeuralNetRegressor
from skorch.callbacks import LRScheduler


#Here fit the models for variable 'bio1' only


class PaleoNN(torch.nn.Module):
	def __init__(self, units: int):
		super().__init__()
		input_features = 599
		layers = [nn.Linear(input_features, units), nn.ReLU(), nn.BatchNorm1d(units)]
		while units > 1:
			prev_units = units
			units = int(units / 2)
			layers += [nn.Linear(prev_units, units), nn.ReLU(), nn.BatchNorm1d(units)]
		layers += [nn.Linear(units, 1)]
		self.model = nn.Sequential(*layers)

	def forward(self, x):
		return self.model(x)



def main(target_variable, root: Path):
	target_variables = ['AI', 'bio1', 'bio2', 'bio3',
								'bio4', 'bio5', 'bio6', 'bio7', 'bio8', 'bio9', 'bio10', 'bio11',
								'bio12', 'bio13', 'bio14', 'bio15', 'bio16', 'bio17', 'bio18',
								'bio19']
	
	folds_dataframes = [
		pd.read_csv(root / f"fold{i+1}.csv", sep=";", decimal=",").drop(columns=['X', 'Y']) for i in range(5)
	]
	test = pd.read_csv(root / "test_set.csv", sep=";", decimal=",").drop(columns=['X', 'Y'])
	folds = [
		[
			fold.drop(columns=['AI', 'bio1', 'bio2', 'bio3',
								'bio4', 'bio5', 'bio6', 'bio7', 'bio8', 'bio9', 'bio10', 'bio11',
								'bio12', 'bio13', 'bio14', 'bio15', 'bio16', 'bio17', 'bio18',
								'bio19']
					).values.astype(np.float32),
			fold[target_variable].values.astype(np.float32)[:,None]
		] for fold in folds_dataframes
	]
	Xtest = test.drop(columns=['AI', 'bio1', 'bio2', 'bio3','bio4', 'bio5', 'bio6', 'bio7', 'bio8', 'bio9', 'bio10', 
							'bio11','bio12', 'bio13', 'bio14', 'bio15', 'bio16', 'bio17', 'bio18','bio19']).values.astype(np.float32)
	ytest =  test[target_variable].values.astype(np.float32)[:,None]  # Target variable for the test data
		
	# Splits for training and validation data
	train_val_splits = folds

	# Concatenated features and target variable for training and validation
	X_train_val = np.concatenate([x for x,_ in train_val_splits])
	y_train_val = np.concatenate([y for _,y in train_val_splits]) 

	# Start indices of each split
	split_start_indices = np.cumsum([0] + [len(x) for x,_ in train_val_splits])
	split_indices = np.concatenate([ i*np.ones(y.shape) for i, (_, y) in enumerate(train_val_splits) ]) # split_indices[i] = 0 means sample i is part of fold0, and so on

	# Predefined cross-validation split
	cv = PredefinedSplit(split_indices)

	# Grid search with cross-validation
	estimator = NeuralNetRegressor(
		PaleoNN,
		criterion=nn.MSELoss(),
		optimizer=torch.optim.Adam,
		optimizer__lr=1e-3,
		device="cuda",
		train_split=None,
		max_epochs = 2,
		batch_size = 4096,
		callbacks=[
			('lr_scheduler',LRScheduler(torch.optim.lr_scheduler.ReduceLROnPlateau, mode='min', factor=0.2, patience=10, threshold=0.01, threshold_mode='rel', cooldown=25, min_lr=5e-6, eps=1e-08, verbose=False)),
		],
		verbose=4
	)
	search_model = GridSearchCV(estimator=estimator, param_grid={'module__units': [200, 500, 1000]}, cv=cv, refit=True, verbose=4) # Since refit=True, best model will be refit on the whole training+validation data 

	# Fit the model on training and validation data
	search_model.fit(X_train_val, y_train_val) 

	# Make predictions on the test data
	y_pred = search_model.predict(Xtest)
	test_metrics = {
		"target": target_variable,
		"best_params": search_model.best_params_,
		"r2": float(r2_score(ytest, y_pred)),
		"rmse": float(sqrt(mean_squared_error(ytest, y_pred))),
		"mae": float(mean_absolute_error(ytest, y_pred)),
		"ajusted-r2": float(1 - ((1 - r2_score(ytest, y_pred)) * (len(ytest) - 1) / (len(ytest) - Xtest.shape[1] - 1)))
	}
	print(f"Best params: {search_model.best_params_}")
	print(f"Test metrics: {test_metrics}")

	print(test_metrics)

	import json
	with (Path.cwd().parent / "out" / f"result-{target_variable}.json").open("w") as f:
		json.dump(test_metrics, f)


if __name__ == "__main__":
	from argparse import ArgumentParser
	parser = ArgumentParser()
	parser.add_argument("--target", type=str, help="Target variable (eg. 'AI')", required=True)
	args = parser.parse_args()
	main(args.target, Path(__file__).resolve().parent / "data")


#target_variables = ["bio1"] #, 'bio2', 'bio3', 'bio4', 'bio5', 'bio6', 'bio7', 'bio8', 'bio9', 'bio10', 'bio11', 'bio12', 'bio13', 'bio14', 'bio15', 'bio16', 'bio17', 'bio18', 'bio19'] # Can test more target variables here
#for target_variable in target_variables:
#  main(target_variable)





